;;;
;;; bitio.lisp : ビット入出力
;;;
;;; Copyright (c) 2023 Makoto Hiroi
;;;
;;; Released under the MIT license
;;; https://opensource.org/license/mit/
;;;
(provide :bitio)
(defpackage :bitio (:use :cl))
(in-package :bitio)
(export '(bit-io
	  bit-io-p
	  call-with-bit-input-file
	  call-with-byte-input-file
          call-with-bit-output-file
	  call-with-byte-output-file
          getbit
	  getbits
	  putbit
	  putbits
          alpha-encode
	  alpha-decode
	  gamma-encode
	  gamma-decode
          delta-encode
	  delta-decode
	  cbt-encode
	  cbt-decode
          rice-encode
	  rice-decode))

;;; 型の定義
(defstruct bit-io direction file buff cnt)

;;; バイト入力用ファイルオープン
(defun call-with-byte-input-file (filename proc)
  (with-open-file (in filename
                      :direction :input
                      :element-type 'unsigned-byte)
    (funcall proc in)))

;;; バイト出力用ファイルオープン
(defun call-with-byte-output-file (filename proc)
  (with-open-file (out filename
                       :direction :output
                       :if-exists :rename-and-delete
                       :element-type 'unsigned-byte)
    (funcall proc out)))

;;; ビット入力用ファイルオープン
(defun call-with-bit-input-file (filename proc)
  (call-with-byte-input-file
   filename
   (lambda (in)
     (funcall proc (make-bit-io :direction :input
                                :file in
                                :cnt 0)))))

;;; ビット出力用ファイルオープン
(defun call-with-bit-output-file (filename proc)
  (call-with-byte-output-file
   filename
   (lambda (out)
     (let ((bs (make-bit-io :direction :output
                            :file out
                            :buff 0
                            :cnt 8)))
       (funcall proc bs)
       (if (< (bit-io-cnt bs) 8)
           (write-byte (bit-io-buff bs) out))))))

;;; 1 ビット入力
(defun getbit (bs)
  (decf (bit-io-cnt bs))
  (when (minusp (bit-io-cnt bs))
    (setf (bit-io-buff bs)
          (read-byte (bit-io-file bs) nil))
    (if (null (bit-io-buff bs))
        (return-from getbit nil))
    (setf (bit-io-cnt bs) 7))
  (if (logbitp (bit-io-cnt bs) (bit-io-buff bs)) 1 0))

;;; 1 ビット出力
(defun putbit (bs val)
  (decf (bit-io-cnt bs))
  (when (plusp val)
    (setf (bit-io-buff bs)
          (logior (bit-io-buff bs) (ash 1 (bit-io-cnt bs)))))
  (when (zerop (bit-io-cnt bs))
    (write-byte (bit-io-buff bs) (bit-io-file bs))
    (setf (bit-io-buff bs) 0
          (bit-io-cnt bs) 8)))

;;; n ビット入力
(defun getbits (bs n)
  (do ((pat (ash 1 (1- n)) (ash pat -1))
       (val 0))
      ((zerop pat) val)
    (case (getbit bs)
      (1 (setf val (logior val pat)))
      (nil (return)))))

;;; n ビット出力
(defun putbits (bs n x)
  (do ((pat (ash 1 (1- n)) (ash pat -1)))
      ((zerop pat))
      (putbit bs (logand x pat))))

;;;
;;; 整数の符号化
;;;

;;; α符号
(defun alpha-encode (bs n)
  (putbits bs n 0)
  (putbit bs 1))

(defun alpha-decode (bs)
  (do ((n 0 (1+ n)))
      ((plusp (getbit bs)) n)))

;;; γ符号
(defun gamma-encode (bs n)
  (do ((n1 0 (1+ n1))
       (n2 (ash (1+ n) -1) (ash n2 -1)))
      ((zerop n2)
       (alpha-encode bs n1)
       (if (plusp n1) (putbits bs n1 (1+ n))))))

(defun gamma-decode (bs)
  (let ((n1 (alpha-decode bs)))
    (if (zerop n1)
        0
      (+ (ash 1 n1) (getbits bs n1) -1))))

;;; δ符号
(defun delta-encode (bs n)
  (do ((n1 0 (1+ n1))
       (n2 (ash (1+ n) -1) (ash n2 -1)))
      ((zerop n2)
       (gamma-encode bs n1)
       (if (plusp n1) (putbits bs n1 (1+ n))))))

(defun delta-decode (bs)
  (let ((n1 (gamma-decode bs)))
    (if (zerop n1)
        0
      (+ (ash 1 n1) (getbits bs n1) -1))))

;;; CBT 符号
(defun cbt-encode (bs n m k)
  (let ((limit (- (ash 1 k) m)))
    (if (< n limit)
        (putbits bs (1- k) n)
      (putbits bs k (+ n limit)))))

(defun cbt-decode (bs m k)
  (let ((limit (- (ash 1 k) m))
        (n (getbits bs (1- k))))
    (if (< n limit)
        n
      (+ (ash n 1) (getbit bs) (- limit)))))

;;; Rice 符号
(defun rice-encode (bs n k)
  (alpha-encode bs (ash n (- k)))
  (putbits bs k n))

(defun rice-decode (bs k)
  (let ((n (alpha-decode bs)))
    (+ (ash n k) (getbits bs k))))

