;;;
;;; bitree.lisp : binary indexed tree
;;;
;;; Copyright (c) 2023 Makoto Hiroi
;;;
;;; Released under the MIT license
;;; https://opensource.org/license/mit/
;;;
(provide :bitree)
(defpackage :bitree (:use :cl))
(in-package :bitree)
(export '(bitree
	  bitree-p
	  make-bitree
	  bitree-initialize
	  bitree-get
	  bitree-sum
	  bitree-add
          bitree-total
	  bitree-find
	  bitree-half
	  bitree-clear))

;;; 中央値 (ルート) を求める
(defun get-mid (size)
  (do ((mid 1 (ash mid 1))
       (limit (ash size -1)))
      ((<= limit mid) mid)))

;;; 構造体の定義
(defstruct (bitree
            (:constructor make-bitree
             (size
              &aux
              (table (make-array size :initial-element 0))
              (mid (get-mid size))
              (total 0))))
  size table mid total)

;;; 初期化
(defun bitree-initialize (code-size)
  (let ((bt (make-bitree code-size)))
    (dotimes (x code-size bt) (bitree-add bt x 1))))

;;; c の値 (出現頻度) を求める
(defun bitree-get (bt c)
  (let ((n (aref (bitree-table bt) c)))
    (if (or (zerop c) (oddp c))
        n
      (do ((p (logand c (1- c)))
           (x (1- c) (logand x (1- x))))
          ((= x p) n)
        (decf n (aref (bitree-table bt) x))))))

;;; 0 - c-1 までの累積度数を求める
(defun bitree-sum (bt c)
  (if (zerop c)
      0
    (let ((n (aref (bitree-table bt) 0)))
      ;; c - 1 までの頻度を加算する
      (do ((x (1- c) (logand x (1- x))))
          ((zerop x) n)
        (incf n (aref (bitree-table bt) x))))))

;;; c の値に inc を加算する
(defun bitree-add (bt c inc)
  (if (zerop c)
      (incf (aref (bitree-table bt) c) inc)
    (do ()
        ((<= (bitree-size bt) c))
      (incf (aref (bitree-table bt) c) inc)
      (incf c (logand c (- c)))))
  (incf (bitree-total bt) inc))

;;; 探索 (cumul[c] <= val < cumul[c + 1])
(defun bitree-find (bt val)
  (let ((n (aref (bitree-table bt) 0)))
    (if (< val n)
        (values 0 0)
      (do ((h (bitree-mid bt) (ash h -1))
           (c 0))
          ((zerop h) (values (1+ c) n))
        (when (and (< (+ c h) (bitree-size bt))
                   (<= (+ n (aref (bitree-table bt) (+ c h))) val))
          (incf n (aref (bitree-table bt) (+ c h)))
          (incf c h))))))

;;; total が半分になるよう個数を調整
(defun bitree-half (bt)
  (dotimes (x (bitree-size bt))
    (let ((n (ash (bitree-get bt x) -1)))
      (when (plusp n)
        (bitree-add bt x (- n))))))

;;; 空にする
(defun bitree-clear (bt)
  (fill (bitree-table bt) 0)
  (setf (bitree-total bt) 0))

