;;;
;;; combination.lisp : 順列と組み合わせ
;;;
;;; Copyright (c) 2023 Makoto Hiroi
;;;
;;; Released under the MIT license
;;; https://opensource.org/license/mit/
;;;
(provide :combination)
(defpackage :combination (:use :cl))
(in-package :combination)
(export '(factorial
	  permutation-number
	  permutation
	  permutation-set
	  repeated-permutation
	  repeated-permutation-set
	  montmort-number
	  derangement
	  derangement-set
	  combination-number
	  combination
	  combination-set
	  repeated-combination-number
	  repeated-combination
	  repeated-combination-set
	  product
	  product-set
	  fibonacci
	  catalan-number
	  partition-number
	  stirling2-number
	  bell-number
	  ))

;;; 階乗 n!
(defun factorial (n &optional (a 1))
  (if (zerop n)
      a
    (factorial (1- n) (* n a))))

;;;
;;; 順列
;;;

;;; n 個の中らから k 個を選ぶ順列
(defun permutation-number (n k &optional (a 1))
  (if (zerop k)
      a
    (permutation-number (1- n) (1- k) (* a n))))

;;; 高階関数版
(defun permutation (fn n xs &optional (a nil))
  (if (zerop n)
      (funcall fn (reverse a))
    (dolist (x xs)
      (permutation fn (1- n) (remove x xs :count 1) (cons x a)))))
 
;;; リストに格納して返す
(defun permutation-set (n xs)
  (if (zerop n)
      '(())
    (mapcan (lambda (x)
	      (mapcar (lambda (ys) (cons x ys))
		      (permutation-set (1- n) (remove x xs :count 1))))
	    xs)))

;;; 重複順列 (高階関数版)
(defun repeated-permutation (fn n xs &optional (a nil))
  (if (zerop n)
      (funcall fn (reverse a))
    (dolist (x xs)
      (repeated-permutation fn (1- n) xs (cons x a)))))

;;; 重複順列 (リストに格納して返す)
(defun repeated-permutation-set (n xs)
  (if (zerop n)
      '(())
    (mapcan (lambda (x)
	      (mapcar (lambda (ys) (cons x ys))
		      (repeated-permutation-set (1- n) xs)))
	    xs)))

;;;
;;; 完全順列
;;;

;;; 完全順列の総数 (モンモール数)
(defun montmort-number (n)
  (do ((i 1 (1+ i))
       (a 0)
       (b 1))
      ((>= i n) a)
      (psetq a b
             b (* (1+ i) (+ a b)))))

;;; 高階関数
(defun derangement-sub (fn xs zs a)
  (if (null xs)
      (funcall fn (reverse a))
    (dolist (z zs)
      (unless (eql (car xs) z)
	(derangement-sub fn (cdr xs) (remove z zs :count 1) (cons z a))))))

(defun derangement (fn xs)
  (derangement-sub fn xs xs nil))

;;; リストに格納して返す
(defun derangement-set-sub (xs zs)
  (if (null xs)
      '(())
    (mapcan (lambda (z)
	      (unless (eql (car xs) z)
		(mapcar (lambda (ls) (cons z ls))
			(derangement-set-sub (cdr xs) (remove z zs :count 1)))))
	    zs)))

(defun derangement-set (xs) (derangement-set-sub xs xs))

;;;
;;; 組み合わせ
;;;

;;; 組み合わせの数
(defun combination-number (n r)
  (if (or (= n r) (zerop r))
      1
    (/ (* (combination-number n (1- r)) (1+ (- n r))) r)))

;;; 高階関数版
(defun combination (fn n xs &optional (a nil))
  (cond
   ((zerop n)
    (funcall fn (reverse a)))
   ((null xs) nil)
   (t
    (combination fn (1- n) (cdr xs) (cons (car xs) a))
    (combination fn n (cdr xs) a))))

;;; リストに格納して返す
(defun combination-set (n xs)
  (cond
   ((zerop n) '(()))
   ((null xs) nil)
   (t
    (append
     (mapcar (lambda (ys) (cons (car xs) ys))
	     (combination-set (1- n) (cdr xs)))
     (combination-set n (cdr xs))))))

;;; 重複組み合わせの数
(defun repeated-combination-number (n r) (combination-number (1- (+ n r)) r))

;;; 重複組み合わせ (高階関数版)
(defun repeated-combination (fn n xs &optional (a nil))
  (cond
   ((zerop n)
    (funcall fn (reverse a)))
   ((null xs) nil)
   (t
    (repeated-combination fn (1- n) xs (cons (car xs) a))
    (repeated-combination fn n (cdr xs) a))))

;;; 重複組み合わせ (リストに格納して返す)
(defun repeated-combination-set (n xs)
  (cond
   ((zerop n) '(()))
   ((null xs) nil)
   (t
    (append (mapcar (lambda (ys) (cons (car xs) ys))
		    (repeated-combination-set (1- n) xs))
	    (repeated-combination-set n (cdr xs))))))

;;; 直積集合
(defun product-sub (fn a xss)
  (if (null xss)
      (funcall fn (reverse a))
    (dolist (x (car xss))
      (product-sub fn (cons x a) (cdr xss)))))

(defun product (fn &rest args)
  (product-sub fn nil args))

(defun product-set (&rest args)
  (cond
   ((null args) (list '()))
   ((null (cdr args))
    (mapcar (lambda (x) (list x)) (car args)))
   (t
    (mapcan (lambda (x) (mapcar (lambda (ys) (cons x ys))
                                (apply #'product-set (cdr args))))
            (car args)))))

;;;
;;; 数
;;;

;;; フィボナッチ数
(defun fibonacci (n &optional (a 0) (b 1))
  (if (zerop n)
      a
    (fibonacci (1- n) b (+ a b))))

;;; カタラン数
(defun catalan-number (n)
  (/ (combination-number (* n 2) n) (1+ n)))

;;; 分割数 (自然数の分割)
(defun partition-number (n)
  (let ((a (make-array (+ n 1) :initial-element 1)))
    (do ((k 2 (1+ k)))
        ((< n k) (aref a n))
      (do ((m k (1+ m)))
          ((< n m))
        (incf (aref a m) (aref a (- m k)))))))

;;; 第 2 種スターリング数
(defun stirling2-sub (n)
  (let ((xs (list 0 1)))
    (dotimes (i (1- n) xs)
      (do ((j 0 (1+ j))
	   (zs xs (cdr zs))
	   (ys (list 0)))
	  ((null (cdr zs))
	   (setq ys (cons 1 ys)
		 xs (nreverse ys)))
	  (push (+ (car zs) (* (1+ j) (cadr zs))) ys)))))

(defun stirling2-number (n k)
  (if (= n k 0)
      1
    (elt (stirling2-sub n) k)))

;;; ベル数
(defun bell-number (n)
  (if (or (zerop n) (= n 1))
      1
    (reduce #'+ (stirling2-sub n))))

