;;;
;;; deque.lisp : 両端キューと双方向リスト
;;;
;;; Copyright (C) 2023 Makoto Hiroi
;;;
;;; Released under the MIT license
;;; https://opensource.org/license/mit/
;;;
(provide :deque)
(defpackage :deque (:use :cl))
(in-package :deque)
(export '(dlist dlist-p make-dlist dlist-emptyp dlist-count dlist-clear
	  dlist-nth dlist-set dlist-insert-at dlist-delete-at
	  dlist-exists dlist-find dlist-find-if dlist-find-if-not
	  dlist-delete dlist-delete-if dlist-delete-if-not
          dlist-fold list->dlist dlist->list
	  dlist-mapc dlist-iterator
	  deque deque-p make-deque
	  push-front push-back
	  pop-front pop-back
	  peek-front peek-back
	  deque-count deque-clear deque-emptyp))

;;; セルの定義
(defstruct cell (item nil) (next nil) (prev nil))

;;; 空リストの生成
(defun make-empty ()
  (let ((cp (make-cell)))
    (setf (cell-next cp) cp
          (cell-prev cp) cp)
    cp))

;;; 双方向リストクラスの定義
(defstruct dlist (top (make-empty)) (size 0))

;;; n 番目のセルを求める (操作用関数)
(defun cell-nth (d n iter)
  (do ((i -1 (1+ i))
       (cp (dlist-top d) (funcall iter cp)))
      ((= i n) cp)
    (if (and (<= 0 i) (eq (dlist-top d) cp))
        (error "cell-nth --- oops!"))))

;;; 参照
(defun dlist-nth (d n &key (from-end nil))
  (cell-item (cell-nth d n (if from-end #'cell-prev #'cell-next))))

;;; 書き換え
(defun dlist-set (d n value &key (from-end nil))
  (setf (cell-item (cell-nth d n (if from-end #'cell-prev #'cell-next)))
        value))

;;;
;;; 探索
;;;
(defun dlist-exists (d x &key (key #'identity) (test #'eql))
  (do ((cp (cell-next (dlist-top d)) (cell-next cp)))
      ((eq (dlist-top d) cp))
      (when (funcall test (funcall key (cell-item cp)) x)
	(return t))))

(defun dlist-find (d x &key (key #'identity) (test #'eql) (from-end nil))
  (let ((iter (if from-end #'cell-prev #'cell-next)))
    (do ((cp (funcall iter (dlist-top d)) (funcall iter cp)))
        ((eq (dlist-top d) cp))
	(when (funcall test (funcall key (cell-item cp)) x)
	  (return (cell-item cp))))))
  
(defun dlist-find-if (d pred &key (key #'identity) (from-end nil))
  (let ((iter (if from-end #'cell-prev #'cell-next)))
    (do ((cp (funcall iter (dlist-top d)) (funcall iter cp)))
        ((eq (dlist-top d) cp))
	(when (funcall pred (funcall key (cell-item cp)))
	  (return (cell-item cp))))))
  
(defun dlist-find-if-not (d pred &key (key #'identity) (from-end nil))
  (let ((iter (if from-end #'cell-prev #'cell-next)))
    (do ((cp (funcall iter (dlist-top d)) (funcall iter cp)))
        ((eq (dlist-top d) cp))
	(unless (funcall pred (funcall key (cell-item cp)))
	  (return (cell-item cp))))))
  

;;; セルの挿入
;;; p - next -> cp - next -> q
(defun cell-insert (p cp q)
  (setf (cell-next cp) q
        (cell-prev cp) p
        (cell-prev q) cp
        (cell-next p) cp))

;;; 挿入
(defun dlist-insert-at (d n value &key (from-end nil))
  (let* ((iter (if from-end #'cell-prev #'cell-next))
         (p (cell-nth d (1- n) iter))
         (q (funcall iter p))
         (cp (make-cell :item value)))
    (if from-end
        (cell-insert q cp p)
      (cell-insert p cp q))
    (incf (dlist-size d))
    value))

;;; セルの削除
;;; p - next -> [cp] - next -> q
(defun cell-delete (p q)
  (setf (cell-next p) q
        (cell-prev q) p))

;;; 削除
(defun dlist-delete-at (d n &key (from-end nil))
  (let* ((iter (if from-end #'cell-prev #'cell-next))
         (p  (cell-nth d (1- n) iter))
         (cp (funcall iter p))
         (q  (funcall iter cp)))
    (if from-end (cell-delete q p) (cell-delete p q))
    (decf (dlist-size d))
    (cell-item cp)))

(defun dlist-delete (d x &key (key #'identity) (test #'eql) (from-end nil) (count 0))
  (let ((iter (if from-end #'cell-prev #'cell-next)))
    (do ((cp (funcall iter (dlist-top d)) (funcall iter cp)))
        ((eq (dlist-top d) cp) d)
	(when (funcall test (funcall key (cell-item cp)) x)
	  (cell-delete (cell-prev cp) (cell-next cp))
	  (decf (dlist-size d))
	  (decf count)
	  (when (zerop count)
	    (return d))))))

(defun dlist-delete-if (d pred &key (key #'identity) (from-end nil) (count 0))
  (let ((iter (if from-end #'cell-prev #'cell-next)))
    (do ((cp (funcall iter (dlist-top d)) (funcall iter cp)))
        ((eq (dlist-top d) cp) d)
	(when (funcall pred (funcall key (cell-item cp)))
	  (cell-delete (cell-prev cp) (cell-next cp))
	  (decf (dlist-size d))
	  (decf count)
	  (when (zerop count)
	    (return d))))))

(defun dlist-delete-if-not (d pred &key (key #'identity) (from-end nil) (count 0))
  (let ((iter (if from-end #'cell-prev #'cell-next)))
    (do ((cp (funcall iter (dlist-top d)) (funcall iter cp)))
        ((eq (dlist-top d) cp) d)
	(unless (funcall pred (funcall key (cell-item cp)))
	  (cell-delete (cell-prev cp) (cell-next cp))
	  (decf (dlist-size d))
	  (decf count)
	  (when (zerop count)
	    (return d))))))

;;; 畳み込み
(defun dlist-fold (d func init &key from-end)
  (let ((iter (if from-end #'cell-prev #'cell-next)))
    (do ((cp (funcall iter (dlist-top d)) (funcall iter cp))
         (a init))
        ((eq cp (dlist-top d)) a)
      (setq a (if from-end
                  (funcall func (cell-item cp) a)
                (funcall func a (cell-item cp)))))))

;;; サイズ
(defun dlist-count (d) (dlist-size d))

;;; クリア
(defun dlist-clear (d)
  (let ((cp (dlist-top d)))
    (setf (cell-next cp) cp
          (cell-prev cp) cp
	  (dlist-size d) 0)))

;;; 空リストか？
(defun dlist-emptyp (d)
  (let ((cp (dlist-top d)))
    (eq cp (cell-next cp))))

;;; リストを双方向リストに変換
(defun list->dlist (xs)
  (let ((d (make-dlist)))
    (dolist (x xs d)
      (dlist-insert-at d 0 x :from-end t))))

;;; 双方向リストをリストに変換
(defun dlist->list (d)
  (dlist-fold d (lambda (x y) (cons x y)) nil :from-end t))

;;; 巡回
(defun dlist-mapc (d func &key (from-end nil))
  (let ((iter (if from-end #'cell-prev #'cell-next)))
    (do ((cp (funcall iter (dlist-top d)) (funcall iter cp)))
        ((eq (dlist-top d) cp))
        (funcall func (cell-item cp)))))

;;; イテレータの生成
(defun dlist-iterator (d &key (from-end nil))
  (let* ((iter (if from-end #'cell-prev #'cell-next))
         (cp   (funcall iter (dlist-top d))))
    (lambda ()
      (if (eq (dlist-top d) cp)
          (values nil nil)
        (multiple-value-prog1
            (values (cell-item cp) t)
          (setq cp (funcall iter cp)))))))

;;; 表示
(defmethod print-object ((x dlist) stream)
  (format stream "#<dlist: ~S>" (dlist->list x)))

;;;
;;; 両端キュー
;;;

;;; 定義
(defstruct deque (top (make-dlist)))

;;; データの追加
(defun push-front (d value)
  (dlist-insert-at (deque-top d) 0 value))

(defun push-back (d value)
  (dlist-insert-at (deque-top d) 0 value :from-end t))

;;; データの取り出し
(defun pop-front (d)
  (dlist-delete-at (deque-top d) 0))

(defun pop-back (d)
  (dlist-delete-at (deque-top d) 0 :from-end t))

;;; データの参照
(defun peek-front (d)
  (dlist-nth (deque-top d) 0))

(defun peek-back (d)
  (dlist-nth (deque-top d) 0 :from-end t))

;;; 要素数を求める
(defun deque-count (q) (dlist-count (deque-top q)))

;;; ディーキューを空にする
(defun deque-clear (q) (dlist-clear (deque-top q)))

;;; ディーキューは空か？
(defun deque-emptyp (q) (dlist-emptyp (deque-top q)))

