;;;
;;; heap.lisp : ヒープ
;;;
;;; Copyright (C) 2023 Makoto Hiroi
;;;
;;; Released under the MIT license
;;; https://opensource.org/license/mit/
;;;
(provide :heap)
(defpackage :heap (:use :cl))
(in-package :heap)
(export '(make-heap
	  heap-p
	  make-leftist-heap
	  leftist-heap-p
	  make-skew-heap
	  skew-heap-p
	  heap-push
	  heap-pop
	  heap-peek
	  heap-copy
	  heap-merge
          heap-count
	  heap-clear
	  heap-emptyp))

;;; 総称関数の定義
(defgeneric heap-push (h x))
(defgeneric heap-pop (h))
(defgeneric heap-peek (h))
(defgeneric heap-copy (h))
(defgeneric heap-merge (h1 h2))
(defgeneric heap-count (h))
(defgeneric heap-clear (h))
(defgeneric heap-emptyp (h))

;;; 定義
(defstruct heap
  (buff (make-array 256 :fill-pointer 0 :adjustable t))
  (key  #'identity)
  (elt> #'>))

;;;
;;; 作業用関数
;;;

;;; 要素の比較
(defun elt> (h x y)
  (funcall (heap-elt> h)
           (funcall (heap-key h) (aref (heap-buff h) x))
           (funcall (heap-key h) (aref (heap-buff h) y))))

;;; 要素の交換
(defun swap (buff x y)
  (psetf (aref buff x) (aref buff y)
         (aref buff y) (aref buff x)))

;;; ヒープの構築
(defun upheap (h n)
  (do ((p (floor (1- n) 2) (floor (1- n) 2)))
      ((or (minusp p) (not (elt> h p n))))
      (swap (heap-buff h) p n)
      (setf n p)))

;;; ヒープの再構築
(defun downheap (h n nums)
  (do ((c (+ (* n 2) 1) (+ (* n 2) 1)))
      ((>= c nums))
      (when (and (< (1+ c) nums) (elt> h c (1+ c)))
        (incf c))
      (when (not (elt> h n c))
        (return))
      (swap (heap-buff h) n c)
      (setf n c)))

;;;
;;; 操作関数の定義
;;;

;;; 空か
(defmethod heap-emptyp ((h heap))
  (zerop (fill-pointer (heap-buff h))))

;;; 要素数
(defmethod heap-count ((h heap))
  (fill-pointer (heap-buff h)))

;;; クリア
(defmethod heap-clear ((h heap))
  (setf (fill-pointer (heap-buff h)) 0))

;;; データの追加
(defmethod heap-push ((h heap) x)
  (vector-push-extend x (heap-buff h))
  (upheap h (- (heap-count h) 1)))

;;; 先頭データの参照
(defmethod heap-peek ((h heap))
  (if (heap-emptyp h)
      (error "heap : heap is empty")
    (aref (heap-buff h) 0)))

;;; データの取り出し
(defmethod heap-pop ((h heap))
  (prog1
      (heap-peek h)
    (unless (heap-emptyp h)
      (let ((buff (heap-buff h)))
        (setf (aref buff 0) (vector-pop buff))
        (downheap h 0 (fill-pointer buff))))))

;;; コピー
(defmethod heap-copy ((h heap))
  (let ((newheap (make-heap ::key (heap-key h) :elt> (heap-elt> h))))
    (dotimes (i (heap-count h) newheap)
      (heap-push newheap (aref (heap-buff h) i)))))

;;; ヒープのマージ (遅い)
(defmethod heap-merge ((h1 heap) (h2 heap))
  (unless (and (eq (heap-key h1) (heap-key h2))
	       (eq (heap-elt> h1) (heap-elt> h2)))
    (error "heap-merge, can not merge"))
  (let ((h3 (heap-copy h1)))
    (dotimes (i (heap-count h2) h3)
      (heap-push h3 (aref (heap-buff h2) i)))))

;;;
;;; Leftist Heap
;;;

;;; 節の定義, 終端は nil
(defstruct lnode item rank left right)

;;; マージ
(defun merge-lnode-sub (x hs1 hs2)
  (let ((r1 (if (null hs1) 0 (lnode-rank hs1)))
	(r2 (if (null hs2) 0 (lnode-rank hs2))))
    (if (>= r1 r2)
	(make-lnode :item x :rank (1+ r2) :left hs1 :right hs2)
      (make-lnode :item x :rank (1+ r1) :left hs2 :right hs1))))

(defun merge-lnode (hs1 hs2 key elt<)
  (cond
   ((null hs1) hs2)
   ((null hs2) hs1)
   (t
    (let ((x (lnode-item hs1))
	  (y (lnode-item hs2)))
      (if (funcall elt< (funcall key x) (funcall key y))
	  (merge-lnode-sub x (lnode-left hs1) (merge-lnode (lnode-right hs1) hs2 key elt<))
	(merge-lnode-sub y (lnode-left hs2) (merge-lnode (lnode-right hs2) hs1 key elt<)))))))

;;; 型の定義
(defstruct leftist-heap
  (root nil) (size 0) (key #'identity) (elt< #'<))

;;; ヒープは空か
(defmethod heap-emptyp ((h leftist-heap))
  (null (leftist-heap-root h)))

;;; 要素数を求める
(defmethod heap-count ((h leftist-heap))
  (leftist-heap-size h))

;;; ヒープを空にする
(defmethod heap-clear ((h leftist-heap))
  (setf (leftist-heap-root h) nil
	(leftist-heap-size h) 0))

;;; コピー
(defmethod heap-copy ((h leftist-heap)) (copy-leftist-heap h))

;;; 挿入
(defmethod heap-push ((h leftist-heap) x)
  (setf (leftist-heap-root h)
	(merge-lnode (make-lnode :item x :rank 1 :left nil :right nil)
		     (leftist-heap-root h)
		     (leftist-heap-key h)
		     (leftist-heap-elt< h)))
  (incf (leftist-heap-size h)))

;;; 最小値を取り出す
(defmethod heap-pop ((h leftist-heap))
  (when (heap-emptyp h)
    (error "Leftist Heap is empty"))
  (let ((root (leftist-heap-root h)))
    (prog1
	(lnode-item root)
      (setf (leftist-heap-root h)
	    (merge-lnode (lnode-left root)
			 (lnode-right root)
			 (leftist-heap-key h)
			 (leftist-heap-elt< h)))
      (decf (leftist-heap-size h)))))

;;; 最小値を求める
(defmethod heap-peek ((h leftist-heap))
  (when (heap-emptyp h)
    (error "Leftist Heap is empty"))
  (lnode-item (leftist-heap-root h)))

;;; ヒープのマージ
(defmethod heap-merge ((h1 leftist-heap) (h2 leftist-heap))
  (let ((key (leftist-heap-key h1))
	(elt< (leftist-heap-elt< h1)))
    (unless (and (eq (leftist-heap-key h2) key)
		 (eq (leftist-heap-elt< h2) elt<))
      (error "heap-merge, can not merge"))
    (let ((h3 (make-leftist-heap :key key :elt< elt<)))
      (setf (leftist-heap-root h3)
	    (merge-lnode (leftist-heap-root h1)
			 (leftist-heap-root h2)
			 key
			 elt<)
            (leftist-heap-size h3)
	    (+ (heap-count h1) (heap-count h2)))
      h3)))

;;;
;;; Skew Heap
;;;

;;; 節の定義, 終端は nil
(defstruct snode item left right)

;;; マージ
(defun merge-snode (hs1 hs2 key elt<)
  (cond
   ((null hs2) hs1)
   ((null hs1) hs2)
   (t
    (let ((x (snode-item hs1))
	  (y (snode-item hs2)))
      (if (funcall elt< (funcall key x) (funcall key y))
	  (make-snode :item x
		      :left (merge-snode (snode-right hs1) hs2 key elt<)
		      :right (snode-left hs1))
	(make-snode :item y
		    :left (merge-snode (snode-right hs2) hs1 key elt<)
		    :right (snode-left hs2)))))))

;;; 型の定義
(defstruct skew-heap
  (root nil) (size 0) (key #'identity) (elt< #'<))

;;; キューは空か？
(defmethod heap-emptyp ((h skew-heap))
  (null (skew-heap-root h)))

;;; 要素数を求める
(defmethod heap-count ((h skew-heap))
  (skew-heap-size h))

;;; キューを空にする
(defmethod heap-clear ((h skew-heap))
  (setf (skew-heap-root h) nil
	(skew-heap-size h) 0))

;;; コピー
(defmethod heap-copy ((h skew-heap)) (copy-skew-heap h))

;;; 挿入
(defmethod heap-push ((h skew-heap) x)
  (setf (skew-heap-root h)
	(merge-snode (make-snode :item x :left nil :right nil)
		     (skew-heap-root h)
		     (skew-heap-key h)
		     (skew-heap-elt< h)))
  (incf (skew-heap-size h)))

;;; 最小値を取り出す
(defmethod heap-pop ((h skew-heap))
  (when (heap-emptyp h)
    (error "Skew Heap is empty"))
  (let ((root (skew-heap-root h)))
    (prog1
	(snode-item root)
      (setf (skew-heap-root h)
	    (merge-snode (snode-left root)
			 (snode-right root)
			 (skew-heap-key h)
			 (skew-heap-elt< h)))
      (decf (skew-heap-size h)))))

;;; 最小値を求める
(defmethod heap-peek ((h skew-heap))
  (when (heap-emptyp h)
    (error "Skew Heap is empty"))
  (snode-item (skew-heap-root h)))

;;; ヒープのマージ
(defmethod heap-merge ((h1 skew-heap) (h2 skew-heap))
  (let ((key (skew-heap-key h1))
	(elt< (skew-heap-elt< h1)))
    (unless (and (eq (skew-heap-key h2) key)
		 (eq (skew-heap-elt< h2) elt<))
      (error "heap-merge, can not merge"))
    (let ((h3 (make-skew-heap :key key :elt< elt<)))
      (setf (skew-heap-root h3)
	    (merge-snode (skew-heap-root h1)
			 (skew-heap-root h2)
			 key
			 elt<)
            (skew-heap-size h3)
	    (+ (heap-count h1) (heap-count h2)))
      h3)))

