;;;
;;; ntheory.lisp : Number Theory (整数論)
;;;
;;; Copyright (c) 2023-2024 Makoto Hiroi
;;;
;;; Released under the MIT license
;;; https://opensource.org/license/mit/
;;;
(provide :ntheory)
(defpackage :ntheory (:use :cl :combination :lazy))
(in-package :ntheory)
(export '(*primes*
	  *twin-primes*
	  prime-nth
	  prime-count
	  prime-next
	  prime-prev
	  prime-range
	  sieve
	  segmented-sieve
	  modpow
	  fermat-test
	  miller-rabin-test
	  lucas-lehmer-test
	  factorization
	  factors-to-number
	  divisor-count
	  divisor-total
	  divisor-sigma
	  divisors
	  totient
	  mobius
	  order-n
	  primitive-root-p
	  primitive-root
	  digits
	  perfect-number-p
	  abundant-number-p
	  deficient-number-p
	  amicable-numbers-p
	  extended-euclid
	  polygonal-number
	  polygonal-number-p
	  fermat-number
	  binomial-coefficients
	  crt
	  legendre-symbol
	  jacobi-symbol
	  modsqrt
	  cont-frac-reduce
	  cont-frac
	  cont-frac-float
	  cont-frac-root
	  pells-equation
	  repeat-decimal
	  from-repeat-decimal
	  ))

;;; 素数 (遅延ストリーム版)
(declaim (ftype (function (integer list) t) primes-from))
(defvar *primes* (lcons 2 (lcons 3 (lcons 5 (primes-from 7 '#1=(4 2 4 2 4 6 2 6 . #1#))))))

(defun primep (n &optional (ps (lcdr *primes*)))
  (let ((p (lcar ps)))
    (cond
     ((> (* p p) n) t)
     ((zerop (mod n p)) nil)
     (t (primep n (lcdr ps))))))

(defun primes-from (n ks)
  (if (primep n)
      (lcons n (primes-from (+ n (car ks)) (cdr ks)))
    (primes-from (+ n (car ks)) (cdr ks))))

;;; 双子素数
(defvar *twin-primes*
  (lfilter (lambda (xs) (= (apply #'- xs) -2)) (lzip *primes* (lcdr *primes*))))

;;; n 番目の素数を返す
(defun prime-nth (n) (lnth n *primes*))

;;; n までの素数を個数を求める
(defun prime-count (n &optional (ps *primes*) (c 0))
  (if (< n (lcar ps))
      c
    (prime-count n (lcdr ps) (1+ c))))

;;; n の次の素数を求める
(defun prime-next (n &optional (ps *primes*))
  (lcar (ldrop-while (lambda (p) (<= p n)) ps)))

;;; n の前の素数を求める
(defun prime-prev-sub (n &optional (ps *primes*))
  (if (and (< (lcar ps) n)
	   (<= n (lcar (lcdr ps))))
      (lcar ps)
    (prime-prev-sub n (lcdr ps))))

(defun prime-prev (n)
  (when (< 2 n) (prime-prev-sub n)))

;;; 指定した範囲の素数を求める
(defun prime-range (low high)
  (tolist (ltake-while (lambda (p) (<= p high))
		       (ldrop-while (lambda (p) (< p low)) *primes*))))

;;; エラトステネスの篩
(defun sieve (n)
  ;; 先頭から 3, 5, 7, ...
  (do ((ns (make-array (floor n 2) :element-type '(unsigned-byte 1) :initial-element 0))
       (ps (make-array 1 :fill-pointer t :adjustable t :initial-contents '(2)))
       (x 3 (+ x 2)))
      ((> x n) ps)
      (let ((y (floor (- x 3) 2)))
	(when (zerop (sbit ns y))
	  (vector-push-extend x ps)
	  (when (<= (* x x) n)
	    (do ((y (+ y x) (+ y x)))
		((>= y (length ns)))
		(setf (sbit ns y) 1)))))))

;;; 区間ふるい low 以上 high 以下
(defun segmented-sieve (low high)
  (when (< low 2) (setf low 2))
  (let* ((size (1+ (- high low)))
	 (ns (make-array size :element-type '(unsigned-byte 1) :initial-element 0)))
    (map nil
	 (lambda (p)
	   (do* ((s0 (- (* (ceiling low p) p) low))
		 (s1 (if (< p low) s0 (+ s0 p)) (+ s1 p)))
		((>= s1 size))
		(setf (sbit ns s1) 1)))
	 (sieve (floor (sqrt high))))
    (do ((rs (make-array 1 :fill-pointer 0 :adjustable t))
	 (i 0 (1+ i)))
	((>= i size) rs)
	(when (zerop (sbit ns i))
	  (vector-push-extend (+ low i) rs)))))

;;; べき剰余
;;; (mod (expt b e) m) を高速に求める
(defun modpow (b e m)
  (do ((r 1)
       (e e (ash e -1))
       (b b (mod (* b b) m)))
      ((zerop e) r)
      (when (plusp (logand e 1))
	(setf r (mod (* r b) m)))))

;;; フェルマーテスト
(defun fermat-test (p &optional (k 20))
  (cond
   ((< p 2) nil)
   ((= p 2) t)
   ((zerop (logand p 1)) nil)
   (t
    (do ((k k (1- k)))
	((zerop k) t)
	(let ((a (+ (random (- p 2)) 2)))
	  (when (or (/= (gcd p a) 1)
		    (/= (modpow a (1- p) p) 1))
	    (return nil)))))))

;;; ミラーラビンテスト

;;; n を (2 ** c) * d に分解する 
(defun factor2 (n &optional (c 0))
  (if (= (logand n 1) 1)
      (values c n)
    (factor2 (ash n -1) (1+ c))))

(defun miller-rabin-sub (p s d ks)
  (dolist (a ks t)
    (when (>= a p) (return t))
    (unless (= (modpow a d p) 1)
      (dotimes (r s (return-from miller-rabin-sub nil))
	(when (= (modpow a (* (expt 2 r) d) p) (1- p))
	  (return t))))))

(defun miller-rabin-test (p &optional (k 20))
  (cond
   ((< p 2) nil)
   ((= p 2) t)
   ((zerop (logand p 1)) nil)
   (t
    (multiple-value-bind
     (s d)
     (factor2 (1- p))
     (cond
      ((< p 4759123141)
       (miller-rabin-sub p s d '(2 7 61)))
      ((<= p #x10000000000000000)
       (miller-rabin-sub p s d '(2 325 9375 28178 450775 9780504 1795265022)))
      (t
       (do ((ks nil))
	   ((>= (length ks) k) (miller-rabin-sub p s d ks))
	   (pushnew (1+ (random (- p 2))) ks)))))))) 

;;; リュカ-レーマーテスト
(defun lucas-lehmer-test (p)
  (cond
   ((< p 2) nil)
   ((= p 2) t)
   (t
    (do ((m (1- (expt 2 p)))
	 (x 4 (mod (- (* x x) 2) m))
	 (i (- p 2) (1- i)))
	((zerop i) (zerop (mod x m)))))))

;;;
;;; 素因数分解
;;;
(defun factor-sub (n m &optional (c 0))
  (if (zerop (mod n m))
      (factor-sub (/ n m) m (1+ c))
    (values c n)))

;;; end まで試し割り (wheel factorization)
(defun trial (n &key (end 10000))
  (let ((x 2) (a nil))
    (dolist (k '(1 2 2 . #1=(4 2 4 2 4 6 2 6 . #1#)))
      (cond
       ((= n 1)
	(return (values 1 a)))
       ((< n (* x x))
	(return (values 1 (cons (cons n 1) a))))
       ((> x end)
	(return (values n a)))
       (t
	(multiple-value-bind
	 (c m)
	 (factor-sub n x)
	 (when (plusp c)
	   (push (cons x c) a))
	 (setf n m)
	 (incf x k)))))))

;;;
;;; ポラード・ロー素因数分解法 (Pollard's rho algorithm)
;;;
(defun gen (x n c) (mod (+ (* x x) c) n))

;;; ブレントの変形版
(defun find-factor-brent (n c a &optional (limit 1000000))
  (do* ((xi (gen a n c))
        (xj (gen xi n c))
	(m 2048) (k 1)
	(inc 3)	(cnt 0))
       ((>= cnt limit))
       (do ((l 0) (w 1) (xjs xj))
	   ((>= l k))
	   (setf w 1)
	   (dotimes (i (min (- k l) m))
	     (setf xj (gen xj n c)
		   w  (mod (* (abs (- xj xi)) w) n))
	     (incf l))
           (let ((d (gcd w n)))
             (cond
	      ((< 1 d n)
               (return-from find-factor-brent d))
	      ((= d n)
	       (loop
		(setf xjs (gen xjs n c)
		      d (gcd (abs (- xjs xj)) n))
		(when (< 1 d)
		  (return-from find-factor-brent (if (< d n) d))))))))
       (setf xi xj)
       (dotimes (i inc) (setf xj (gen xj n c)))
       (incf cnt k)
       (setf k (* k 2))
       (incf inc k)))

;;;
;;; ポラードの p-1 法
;;;
(defun find-factor-pm1 (n a &optional (limit 1000000))
  (do ((m 4096)
       (k 2))
      ((>= k limit))
      (let ((a1 a) (k1 k) (w 1))
	(setf w 1)
	(dotimes (i m)
	  (setf a (modpow a k n))
	  (when (zerop a)
	    (return-from find-factor-pm1 nil))
	  (setf w (mod (* w (1- a)) n))
	  (incf k))
	(let ((d (gcd w n)))
	  (cond
	   ((< 1 d n)
	    (return-from find-factor-pm1 d))
	   ((= d n)
	    (loop
	     (setf a1 (modpow a1 k1 n))
	     (incf k1)
	     (setf d (gcd (1- a1) n))
	     (when (< 1 d)
	       (return-from find-factor-pm1 (if (< d n) d))))))))))

;;; ロー法 + p-1 法
(defun find-factor (n &optional (limit 1000000))
  (let ((p (find-factor-brent n 1 2 limit)))
    (unless p
      (setf p (find-factor-pm1 n 3 limit)))
    (when p
      (when (miller-rabin-test p)
	(return-from find-factor p))
      (let ((q (/ n p)))
	(when (miller-rabin-test q)
	  (return-from find-factor q)))
      (find-factor p limit))))

;;; 試し割りのあとロー法と p-1 法で素因数分解する
(defun factorization (n &optional (limit 1000000))
  (multiple-value-bind
   (n ps)
   (trial n)
   (loop
    (cond
     ((= n 1)
      (return (values (nreverse ps) 1)))
     ((miller-rabin-test n)
      (return (values (nreverse (cons (cons n 1) ps)) 1)))
     (t
      (let ((p (find-factor n limit)))
	(unless p
	  (return (values (nreverse ps) n)))
	(multiple-value-bind
	 (c m)
	 (factor-sub n p)
	 (push (cons p c) ps)
	 (setf n m))))))))

;;; 素因数分解したリストから数を求める
(defun factors-to-number (xs)
  (reduce (lambda (a x) (* a (expt (car x) (cdr x)))) xs :initial-value 1))

;;;
;;; 約数
;;;

;;; 個数
(defun divisor-count (n)
  (multiple-value-bind
   (ps m)
   (factorization n)
   (when (= m 1)
     (reduce (lambda (a x) (* a (+ 1 (cdr x))))
             ps
             :initial-value 1))))

;;; 合計値
(defun divisor-total (n)
  (multiple-value-bind
   (ps m)
   (factorization n)
   (when (= m 1)
     (reduce (lambda (a x)
	       ;; 等比数列の和 (p ** (q + 1) - 1) // (p - 1)
	       (* a (/ (1- (expt (car x) (1+ (cdr x)))) (1- (car x)))))
	     ps
             :initial-value 1))))

;;; 約数 
(defun divisor-sub (p n &optional (a nil))
  (if (zerop n)
      (cons 1 a)
    (divisor-sub p (- n 1) (cons (expt p n) a))))

(defun divisors (n)
  (let ((x (factorization n)))
    (sort (reduce (lambda (a y)
                    (mapcar (lambda (xs) (apply #'* xs))
			    (product-set (divisor-sub (car y) (cdr y)) a)))
                  (cdr x)
                  :initial-value (divisor-sub (caar x) (cdar x)))
          #'<)))

;;; 約数関数 σ(n), n の約数の k 乗の総和
(defun divisor-sigma (n k)
  (if (zerop k)
      (divisor-count n)
    (multiple-value-bind
     (ps m)
     (factorization n)
     (when (= m 1)
       (reduce
	(lambda (a xs)
	  (* a (/ (1- (expt (car xs) (* k (1+ (cdr xs)))))
		  (1- (expt (car xs) k)))))
	ps
	:initial-value 1)))))

;;; オイラーのトーシェント関数
(defun totient (n)
  (multiple-value-bind
   (xs m)
   (factorization n)
   (when (= m 1)
     (reduce (lambda (a ps) (/ (* a (1- (car ps))) (car ps)))
	     xs
	     :initial-value n))))

;;; メビウス関数
(defun mobius (n)
  (if (= n 1)
      1
    (multiple-value-bind
     (xs m)
     (factorization n)
     (when (= m 1)
       (if (some (lambda (ps) (< 1 (cdr ps))) xs)
	   0
	 (expt -1 (length xs)))))))

;;; 位数を求める
(defun order-n (a n)
  (when (= (gcd a n) 1)
    (let ((d (1- n)))
      (multiple-value-bind
       (xs m)
       (factorization d)
       (when (= m 1)
	 (loop
	  (dolist (ps xs (return-from order-n d))
	    (let ((p (car ps)))
	      (when (and (zerop (mod d p))
			 (= (modpow a (/ d p) n) 1))
		(setf d (/ d p))
		(return))))))))))

;;; 原始根の判定
(defun primitive-root-p (a n)
  (when (= (gcd a n) 1)
    (multiple-value-bind
     (xs m)
     (factorization (1- n))
     (when (= m 1)
       (notany (lambda (ps) (= (modpow a (/ (1- n) (car ps)) n) 1)) xs)))))

;;; 原始根を求める
(defun primitive-root (fn n)
  (multiple-value-bind
   (xs m)
   (factorization (1- n))
   (when (= m 1)
     (do ((a 2 (1+ a)))
	 ((>= a n))
	 (when (every (lambda (ps) (/= (modpow a (/ (1- n) (car ps)) n) 1)) xs)
	   (funcall fn a))))))

;;; 正整数 n を base 進数の各桁ごとに分解
(defun digits (n &optional (base 10) (a nil))
  (if (zerop n)
      a
    (multiple-value-bind
     (p q)
     (truncate n base)
     (digits p base (cons q a)))))

;;; 完全数
(defun perfect-number-p (n)
  (= (- (divisor-total n) n) n))

;;; 過剰数
(defun abundant-number-p (n)
  (> (- (divisor-total n) n) n))

;;; 不足数
(defun deficient-number-p (n)
  (< (- (divisor-total n) n) n))

;;; 友愛数
(defun amicable-numbers-p (m n)
  (and (= (- (divisor-total m) m) n)
       (= (- (divisor-total n) n) m)))

;;; 拡張ユークリッドの互除法
(defun extended-euclid (a b)
  (do ((xs (list a 1 0))
       (ys (list b 0 1)))
      ((zerop (first ys)) xs)
      (multiple-value-bind
       (q z)
       (floor (first xs) (first ys))
       (psetf xs ys
	      ys (list z
		       (- (second xs) (* q (second ys)))
		       (- (third xs) (* q (third ys))))))))

;;; 多角数
(defun polygonal-number (p n)
  (/ (- (* n n (- p 2)) (* (- p 4) n)) 2))

;;; m は p 角数か？
(defun polygonal-number-p (p m)
  (let* ((a (- p 2))
	 (b (- p 4))
	 (c (+ (* b b) (* 8 a m)))
	 (d (isqrt c)))
    (and (= (* d d) c)
	 (zerop (mod (+ b d) (* 2 a))))))

;;; フェルマー数
(defun fermat-number (n)
  (1+ (expt 2 (expt 2 n))))

;;; 二項係数 (パスカルの三角形の n 段目を返す)
(defun binomial-coefficients (n &optional (a '(1)))
  (if (zerop n)
      a
    (binomial-coefficients
     (1- n)
     (maplist (lambda (xs) (if (null (cdr xs)) 1 (+ (car xs) (cadr xs))))
	      (cons 0 a)))))

;;; 中国剰余定理
(defun crt-sub (a m b n)
  (let ((zs (extended-euclid m n)))
    (when (zerop (mod (- b a) (car zs)))
      (let ((s (/ (- b a) (car zs)))
	    (l (lcm m n)))
	(values (mod (+ (* s m (cadr zs)) a) l) l)))))

(defun crt (xs)
  (reduce (lambda (zs ys)
	    (multiple-value-bind
	     (z l)
	     (crt-sub (car zs) (cadr zs) (car ys) (cadr ys))
	     (when (null z) (return-from crt nil))
	     (list z l)))
	  xs))

;;; ルジャンドル記号
(defun legendre-symbol (a n)
  (when (or (= n 2)
	    (not (miller-rabin-test n)))
    (error "n should be an odd prime"))
  (cond
   ((zerop (mod a n)) 0)
   ((= (modpow a (/ (1- n) 2) n) 1) 1)
   (t -1)))

;;; ヤコビ記号
(defun jacobi-symbol (a n)
  (when (or (minusp n) (evenp n))
    (error "n should be an odd positive integer"))
  (setf a (mod a n))
  (do ((k 1))
      ((<= a 0) (if (= n 1) k 0))
      ;; 第２法則
      (loop
       (unless (zerop (mod a 2)) (return))
       (setf a (/ a 2))
       (let ((r (mod n 8)))
	 (when (or (= r 3) (= r 5))
	   (setf k (- k)))))
      ;; 相互法則
      (psetf a n n a)
      (when (and (= (mod a 4) 3) (= (mod n 4) 3))
	(setf k (- k)))
      (setf a (mod a n))))

;;; 平方非剰余を一つ求める
(defun modsqrt-sub (n)
  (do ((i 2 (1+ i)))
      ((>= i n))
      (when (= (legendre-symbol i n) -1)
	(return i))))

;;; 平方剰余を求める
(defun modsqrt (a n)
  (cond
   ((= (legendre-symbol a n) -1) nil)
   ((= (mod n 4) 3)
    (modpow a (/ (1+ n) 4) n))
   ((= (mod n 8) 5)
    (let ((x (modpow a (/ (+ n 3)  8) n)))
      (if (= (modpow x 2 n) a)
	  x
        (mod (* x (modpow 2 (/ (1- n) 4) n)) n))))
   (t
    ;; Tonelli–Shanks のアルゴリズム
    (multiple-value-bind
     (s k)
     (factor2 (1- n))           ; n - 1 = 2^s * k
     (let* ((d (modsqrt-sub n)) ; 平方非剰余を選ぶ
            (a1 (modpow a k n))
            (d1 (modpow d k n))
            (m 0))
       (dotimes (i s)
	 (when (= (modpow (* a1 (expt d1 m)) (expt 2 (- s i 1)) n) (1- n))
	   (incf m (expt 2 i))))
       (mod (* (modpow a (/ (1+ k) 2) n) (modpow d1 (/ m 2) n)) n))))))

;;;
;;; 連分数
;;;

;;; 分数を正則連分数に変換
(defun cont-frac (x)
  (unless (rationalp x)
    (error "must be rational number"))
  (do ((a (numerator x))
       (b (denominator x))
       (xs nil))
      ((<= b 0) (nreverse xs))
      (multiple-value-bind
       (p q)
       (floor a b)
       (push p xs)
       (setf a b b q))))

;;;
;;; 正則連分数を有理数に戻す
;;;

;;; 漸化式の計算
(defun cont-frac-sub (xs a0 a1)
  (cdr (reduce (lambda (a x) (cons (cdr a) (+ (* x (cdr a)) (car a))))
               (cdr xs)
               :initial-value (cons a0 a1))))

(defun cont-frac-reduce (xs)
  (/ (cont-frac-sub xs 1 (car xs)) (cont-frac-sub xs 0 1)))

;;; 実数を正則連分数に展開する
(defun cont-frac-float (x &key (k 8) (e 1e-16))
  (do ((k k (1- k))
       (xs nil))
      ((zerop k) (nreverse xs))
      (multiple-value-bind
       (p q)
       (floor x)
       (push p xs)
       (when (< (abs q) e)
	 (return (nreverse xs)))
       (setf x (/ q)))))

;;; √ m の連分数展開
(defun cont-frac-root (m)
  (let* ((c0 (isqrt m))
	 (cf (list c0)))
    (if (= (* c0 c0) m)
	cf
      (let ((a 1) (b 0) (c c0))
	(loop
	 (setf b (- (* c a) b)
	       a (floor (- m (* b b)) a)
	       c (floor (+ c0 b) a))
	 (push c cf)
	 (when (= a 1)
	   (return (nreverse cf))))))))

;;; ペル方程式 p^2 - m*q^2 = 1
(defun pells-equation (m)
  (let ((cf (butlast (cont-frac-root m))))
    (unless cf
      (error "pells-equation: domain error"))
    (let ((p (cont-frac-sub cf 1 (car cf)))
          (q (cont-frac-sub cf 0 1)))
      (if (= (- (* p p) (* m q q)) 1)
          (values p q)
        (values (+ (* p p) (* m q q)) (* 2 p q))))))

;;;
;;; 循環小数
;;;

;;; 分数を循環小数に変換する
(defun repeat-decimal (x)
  (unless (rationalp x)
    (error "must be rational number"))
  (do ((m (numerator x))
       (n (denominator x))
       (xs nil) (ys nil))
      (nil)
      (multiple-value-bind
       (p q)
       (floor m n)
       (cond
	((zerop q)
	 (push p ys)
	 (return (list (nreverse ys) '(0))))
	(t
	 (let ((i (position q xs)))
	   (push p ys)
	   (when (and i (>= i 0))
	     (return (list (nreverse (subseq ys (1+ i)))
			   (nreverse (subseq ys 0 (1+ i))))))
	   (push q xs)
	   (setq m (* q 10))))))))


;;; 循環小数を分数に直す
(defun from-repeat-decimal (xs ys)
  (let ((p0 (reduce (lambda (a x) (+ (* a 10) x)) xs :initial-value 0))
	(q0 (expt 10 (1- (length xs))))
	(p1 (reduce (lambda (a x) (+ (* a 10) x)) ys :initial-value 0))
	(q1 (1- (expt 10 (length ys)))))
    (/ (+ (* q1 p0) p1) (* q0 q1))))

