;;;
;;; queue.lisp : キュー
;;;
;;; Copyright (c) 2023 Makoto Hiroi
;;;
;;; Released under the MIT license
;;; https://opensource.org/license/mit/
;;;
(provide :queue)
(defpackage :queue (:use :cl))
(in-package :queue)
(export '(make-queue
	  queue-p
	  make-ring-buffer
	  ring-buffer-p
	  enqueue
	  dequeue
	  queue-peek
	  queue-emptyp
	  queue-fullp
	  queue-count
	  queue-clear))

;;; メソッドの定義
(defgeneric enqueue (q x))
(defgeneric deenque (q))
(defgeneric queue-peek (q))
(defgeneric queue-emptyp (q))
(defgeneric queue-fullp (q))
(defgeneric queue-count (q))
(defgeneric queue-clear (q))

;;; キューの定義
(defstruct queue
  (front nil) (rear nil) (size 0))

;;; キューは空か？
(defmethod queue-emptyp ((q queue))
  (zerop (queue-size q)))

;;; キューは満杯か？
(defmethod queue-fullp ((q queue)) nil)

;;; 要素数を返す
(defmethod queue-count ((q queue)) (queue-size q))

;;; キューを空にする
(defmethod queue-clear ((q queue))
  (setf (queue-front q) nil
	(queue-rear  q) nil
	(queue-size  q) 0))

;;; データの挿入
(defmethod enqueue ((q queue) item)
  (let ((new-cell (list item)))
    (if (queue-emptyp q)
        ;; キューは空の状態
        (setf (queue-front q) new-cell)
      ;; 最終セルを書き換える
      (setf (cdr (queue-rear q)) new-cell))
    (setf (queue-rear q) new-cell))
  (incf (queue-size q))
  ;; item を返す
  item)

;;; データを取得
(defmethod dequeue ((q queue))
  (unless (queue-emptyp q)
    (prog1
        (pop (queue-front q))
      (decf (queue-size q))
      (when (queue-emptyp q)
        ;; キューは空になった
        (setf (queue-rear q) nil)))))

;;; 先頭の要素を返す
(defmethod queue-peek ((q queue))
  (unless (queue-emptyp q)
    (car (queue-front q))))

;;;
;;; リングバッファ
;;;

;;; リングバッファの定義
(defstruct (ring-buffer
            (:constructor make-ring-buffer (size &aux (buffer (make-array size)))))
  (front 0) (rear 0) (count 0) size buffer)

;;; キューは空か？
(defmethod queue-emptyp ((q ring-buffer))
  (zerop (ring-buffer-count q)))

;;; キューは満杯か？
(defmethod queue-fullp ((q ring-buffer))
  (= (ring-buffer-count q) (ring-buffer-size q)))

;;; 要素数を返す
(defmethod queue-count ((q ring-buffer))
  (ring-buffer-count q))

;;; 先頭データの参照
(defmethod queue-peek ((q ring-buffer))
  (unless (queue-emptyp q)
    (aref (ring-buffer-buffer q) (ring-buffer-front q))))

;;; キューを空にする
(defmethod queue-clear ((q ring-buffer))
  (setf (ring-buffer-rear  q) 0
        (ring-buffer-front q) 0
        (ring-buffer-count q) 0))

;;; データの挿入
(defmethod enqueue ((q ring-buffer) item)
  (unless (queue-fullp q)
    (setf (aref (ring-buffer-buffer q) (ring-buffer-rear q)) item)
    (incf (ring-buffer-count q))
    (incf (ring-buffer-rear q))
    (when (= (ring-buffer-rear q) (ring-buffer-size q))
      (setf (ring-buffer-rear q) 0))
    item))

;;; データの取り出し
(defmethod dequeue ((q ring-buffer))
  (unless (queue-emptyp q)
    (prog1
        (aref (ring-buffer-buffer q) (ring-buffer-front q))
      (decf (ring-buffer-count q))
      (incf (ring-buffer-front q))
      (when (= (ring-buffer-front q) (ring-buffer-size q))
        (setf (ring-buffer-front q) 0)))))

