;;;
;;; brc.lisp : 二値レンジコーダ
;;;
;;; Copyright (c) 2023 Makoto Hiroi
;;;
;;; Released under the MIT license
;;; https://opensource.org/license/mit/
;;;
(provide :brc)
(defpackage :brc (:use :cl :rangecoder))
(in-package :brc)
(export '(bit-context
	  bit-context-p
	  make-bit-context
	  bit-encode
	  bit-decode
	  initialize-binary-model
	  binary-model-p
	  bm-encode
	  bm-decode
	  initialize-alpha-model
	  alpha-model-p
	  alpha-encode
	  alpha-decode
	  initialize-bits-model
	  bits-model-p
	  bits-encode
	  bits-decode
	  initialize-gamma-model
	  gamma-model-p
	  gamma-encode
	  gamma-decode
	  ))

;;; ビットの出現頻度表
(defstruct bit-context (c0 1) (sum 2))

;;; 出現頻度表の更新
(defun bit-update (bc bit limit inc)
  (incf (bit-context-sum bc) inc)
  (when (zerop bit)
    (incf (bit-context-c0 bc) inc))
  (when (<= limit (bit-context-sum bc))
    (setf (bit-context-c0 bc)
          (logior (ash (bit-context-c0 bc) -1) 1)
          (bit-context-sum bc)
          (logior (ash (bit-context-sum bc) -1) 1))
    (when (<= (bit-context-sum bc) (bit-context-c0 bc))
      (setf (bit-context-sum bc) (1+ (bit-context-c0 bc))))))

;;; ビットの符号化
(defun bit-encode (rc bc bit &key (limit min-range) (inc 1))
  (let* ((temp (floor (range-coder-range rc) (bit-context-sum bc)))
	 (n (* temp (bit-context-c0 bc))))
    (cond ((plusp bit)
           (incf (range-coder-low rc) n)
           (decf (range-coder-range rc) n))
          (t
           (setf (range-coder-range rc) n)))
    (encode-normalize rc)
    (bit-update bc bit limit inc)))

;;; ビットの復号
(defun bit-decode (rc bc &key (limit min-range) (inc 1))
  (let* ((temp (floor (range-coder-range rc) (bit-context-sum bc)))
	 (c0 (bit-context-c0 bc))
         (n (* temp c0))
         (bit 0))
    (cond ((< (floor (range-coder-low rc) temp) c0)
           (setf (range-coder-range rc) n))
          (t
           (setf bit 1)
           (decf (range-coder-low rc) n)
           (decf (range-coder-range rc) n)))
    (decode-normalize rc)
    (bit-update bc bit limit inc)
    bit))

;;;
;;; binary-model
;;;

;;; 初期化
(defun initialize-bit-context-table (size)
  (map-into (make-array size) #'make-bit-context))

;;; データ型
(defstruct binary-model size table)

;;; 初期化
(defun initialize-binary-model (size)
  (make-binary-model :size size :table (initialize-bit-context-table (1- size))))

;;; 符号化
(defun bm-encode (rc bm c &key (limit min-range) (inc 1))
  (labels ((encode-sub (node)
             (when (plusp node)
               (let* ((p (ash (1- node) -1))
                      (bc (aref (binary-model-table bm) p)))
                 (encode-sub p)
                 ;; 奇数は左の子 (1), 偶数は右の子 (0)
                 (bit-encode rc bc (logand node 1) :limit limit :inc inc)))))
    (encode-sub (+ c (binary-model-size bm) -1))))

;;; 復号
(defun bm-decode (rc bm &key (limit min-range) (inc 1))
  (do ((node 0)
       (node-size (1- (binary-model-size bm))))
      ((<= node-size node) (- node node-size))
    (let* ((bc (aref (binary-model-table bm) node))
           (bit (bit-decode rc bc :limit limit :inc inc)))
      (if (plusp bit)
          (setf node (+ (* node 2) 1))
        (setf node (+ (* node 2) 2))))))

;;;
;;; alpha model
;;;

;;; データ型の定義
(defstruct alpha-model size table)

;;; 初期化
(defun initialize-alpha-model (size)
  (make-alpha-model :size (1- size) :table (initialize-bit-context-table (1- size))))

;;; 符号化
(defun alpha-encode (rc am c &key (limit min-range) (inc 1))
  (do ((n 0 (1+ n)))
      ((>= n (alpha-model-size am)))
      (let ((bit (if (= n c) 1 0)))
	(bit-encode rc (aref (alpha-model-table am) n) bit :limit limit :inc inc)
	(when (plusp bit) (return)))))
       
;;; 復号
(defun alpha-decode (rc am &key (limit min-range) (inc 1))
  (do ((c 0 (1+ c)))
      ((>= c (alpha-model-size am)) c)
      (let ((bit (bit-decode rc (aref (alpha-model-table am) c) :limit limit :inc inc)))
	(when (plusp bit) (return c)))))

;;;
;;; ビット列のモデル
;;;
(defstruct bits-model size table)

;;; 初期化
(defun initialize-bits-model (size)
  (make-bits-model :size size :table (initialize-bit-context-table size)))

(defun bits-encode (rc bs c &key (limit min-range) (inc 1))
  (dotimes (x (bits-model-size bs))
    (let ((bit (logand (ash c (- x)) 1)))
      (bit-encode rc (aref (bits-model-table bs) x) bit :limit limit :inc inc))))

(defun bits-decode (rc bs &key (limit min-range) (inc 1))
  (do ((c 0)
       (x 0 (1+ x)))
      ((>= x (bits-model-size bs)) c)
      (let ((bit (bit-decode rc (aref (bits-model-table bs) x) :limit limit :inc inc)))
	(when (plusp bit)
	  (setf c (logior (ash bit x) c))))))

;;;
;;; gamma model
;;;
(defstruct gamma-model size context1 context2)

(defun get-context-num (n)
  (do ((n2 (ash n -1) (ash n2 -1))
       (n1 0 (1+ n1)))
      ((<= n2 0) n1)))

(defun initialize-gamma-model (size)
  (let* ((csize (get-context-num size))
	 (ct1 (initialize-alpha-model (1+ csize)))
	 (ct2 (make-array (1+ csize))))
    (do ((x 1 (1+ x)))
	((> x csize))
	(setf (aref ct2 x) (initialize-bits-model x)))
    (make-gamma-model :size size :context1 ct1 :context2 ct2))) 

(defun gamma-encode (rc gm c &key (limit min-range) (inc 1))
  (let ((n (get-context-num (1+ c))))
    (alpha-encode rc (gamma-model-context1 gm) n :limit limit :inc inc)
    (when (plusp n)
      (bits-encode rc (aref (gamma-model-context2 gm) n) (1+ c) :limit limit :inc inc))))
	
(defun gamma-decode (rc gm &key (limit min-range) (inc 1))
  (let ((n1 (alpha-decode rc (gamma-model-context1 gm) :limit limit :inc inc)))
    (if (zerop n1)
	n1
      (let ((n2 (bits-decode rc (aref (gamma-model-context2 gm) n1) :limit limit :inc inc)))
	(+ (ash 1 n1) n2 -1)))))


