;;;
;;; trie.lisp : トライ、パトリシア、三分木
;;;
;;; Copyright (C) 2023 Makoto Hiroi
;;;
;;; Released under the MIT license
;;; https://opensource.org/license/mit/
;;;
(provide :trie)
(defpackage :trie (:use :cl))
(in-package :trie)
(export '(trie make-trie trie-p
	  patricia make-patricia patricia-p
	  ternary make-ternary ternary-p
	  trie-emptyp trie-count trie-clear
	  trie-exists trie-get trie-set trie-del
	  trie-fold trie-mapc trie-push trie-pop
	  trie-copy trie-common-prefix
	  ))

;;; メソッドの宣言
(defgeneric trie-emptyp (ts))
(defgeneric trie-count (ts))
(defgeneric trie-clear (ts))
(defgeneric trie-exists (ts seq))
(defgeneric trie-get (ts seq))
(defgeneric trie-set (ts seq val))
(defgeneric trie-del (ts seq))
(defgeneric trie-push (ts seq val))
(defgeneric trie-pop (ts seq))
(defgeneric trie-mapc (ts fn))
(defgeneric trie-fold (ts a fn))
(defgeneric trie-copy (ts))
(defgeneric trie-common-prefix (ts seq fn))

;;;
;;; トライ
;;;

;;; 節
(defstruct node item child (value 'none))

;;; トライ
(defstruct trie
  (root (make-node)) (size 0) (elt= #'eql))

;;;
;;; 作業用操作関数
;;;

;;; 葉
(defun leaf-p (nd)
  (and (node-p nd)
       (not (eq (node-value nd) 'none))))

;;; 挿入
(defun node-insert (nd seq i elt=)
  (if (= (length seq) i)
      nd
    (let ((child (find (elt seq i) (node-child nd) :key #'node-item :test elt=)))
      (unless child
	(setf child (make-node :item (elt seq i)))
	(push child (node-child nd)))
      (node-insert child seq (1+ i) elt=))))

;;; 探索
(defun node-search (nd seq i elt=)
  (if (= (length seq) i)
      nd
    (let ((child (find (elt seq i) (node-child nd) :key #'node-item :test elt=)))
      (when child
	(node-search child seq (1+ i) elt=)))))
	  
;;; 巡回
(defun node-traverse (nd func a)
  (let ((a1 (cons (node-item nd) a)))
    (when (leaf-p nd)
      (funcall func (reverse a1) (node-value nd)))
    (dolist (x (node-child nd))
      (node-traverse x func a1))))

;;; 畳み込み
(defun node-fold (nd func seq a)
  (let ((seq1 (cons (node-item nd) seq)))
    (when (leaf-p nd)
      (setf a (funcall func a (reverse seq1) (node-value nd))))
    (dolist (x (node-child nd) a)
      (setf a (node-fold x func seq1 a)))))

;;;
;;; メソッドの定義
;;;

;;; トライは空か
(defmethod trie-emptyp ((ts trie)) (zerop (trie-size ts)))

;;; 要素数を求める
(defmethod trie-count ((ts trie)) (trie-size ts))

;;; クリア
(defmethod trie-clear ((ts trie))
  (setf (node-child (trie-root ts)) nil
	(trie-size ts) 0))

;;; 挿入
(defmethod trie-set ((ts trie) seq val)
  (let ((nd (node-insert (trie-root ts) seq 0 (trie-elt= ts))))
    (cond
     ((leaf-p nd)
      (setf (node-value nd) val)
      nil)
     (t
      (setf (node-value nd) val)
      (incf (trie-size ts))
      t))))

(defmethod trie-push ((ts trie) seq val)
  (let ((nd (node-insert (trie-root ts) seq 0 (trie-elt= ts))))
    (cond
     ((leaf-p nd)
      (setf (node-value nd) (cons val (node-value nd)))
      nil)
     (t
      (setf (node-value nd) (list val))
      (incf (trie-size ts))
      t))))

;;; 探索
(defmethod trie-exists ((ts trie) seq)
  (leaf-p (node-search (trie-root ts) seq 0 (trie-elt= ts))))

(defmethod trie-get ((ts trie) seq)
  (let ((nd (node-search (trie-root ts) seq 0 (trie-elt= ts))))
    (when (leaf-p nd)
      (node-value nd))))

;;; 削除
(defmethod trie-del ((ts trie) seq)
  (let ((nd (node-search (trie-root ts) seq 0 (trie-elt= ts))))
   (when (leaf-p nd)
     (setf (node-value nd) 'none)
     (decf (trie-size ts))
     (when (zerop (trie-size ts))
       (setf (node-child (trie-root ts)) nil))       
     t)))

(defmethod trie-pop ((ts trie) seq)
  (let ((nd (node-search (trie-root ts) seq 0 (trie-elt= ts))))
    (when (and (leaf-p nd)
	       (consp (node-value nd)))
      (pop (node-value nd)))))

;;; 巡回
(defmethod trie-mapc ((ts trie) func)
  (dolist (x (node-child (trie-root ts)))
    (node-traverse x func nil)))

;;; 畳み込み
(defmethod trie-fold ((ts trie) a func)
  (dolist (x (node-child (trie-root ts)) a)
    (setf a (node-fold x func nil a))))

;;; 共通接頭辞の取得
(defmethod trie-common-prefix ((ts trie) seq func)
  (let ((k (coerce seq 'list)))
    (let ((nd (node-search (trie-root ts) seq 0 (trie-elt= ts))))
      (when nd
	(when (leaf-p nd)
	  (funcall func k (node-value nd)))
	(dolist (x (node-child nd))
	  (node-traverse x func (reverse k)))))))

;;; トライのコピー
(defmethod trie-copy ((ts trie))
  (let ((new-ts (make-trie :elt= (trie-elt= ts))))
    (trie-mapc ts (lambda (k v) (trie-set new-ts k v)))
    new-ts))

;;;
;;; patricia tree
;;;

;;; 型の定義
(defstruct patricia
  (root (make-node)) (size 0) (elt= #'eql))

;;; 子の探索
(defun child-search-pat (nd seq si elt=)
  (dolist (x (node-child nd))
    (when (funcall elt= (elt seq si) (elt (node-item x) 0))
      (return x))))

;;; node から最長一致する節を求める
(defun node-longest-match (nd seq elt=)
  (do ((k (length seq))
       (i 0))
      ((>= i k) (values nd i 0))
      (let ((child (child-search-pat nd seq i elt=)))
	(unless child
	  (return (values nd i 0)))
	(do* ((j 1)
	      (s1 (node-item child))
	      (k1 (length s1)))
	     ((>= j k1) (incf i k1))
	     (when (or (>= (+ i j) k)
		       (not (funcall elt= (elt seq (+ i j)) (elt s1 j))))
	       ;; child の途中まで一致
	       (return-from node-longest-match (values child (+ i j) j)))
	     (incf j))
	(setf nd child))))

;;; 新しい節と葉を追加する
;;; nd <- new-node + val
(defun insert-new-node-leaf (nd seq val)
  (let ((new-node (make-node :item seq :value val)))
    (push new-node (node-child nd))
    t))

;;; 節を分割する
;;; nd -> (nd - child)
(defun divide-node (nd sub-match)
  (let ((child (make-node :item (subseq (node-item nd) sub-match))))
    (setf (node-child child) (node-child nd)
	  (node-value child) (node-value nd)
	  (node-child nd) (list child)
	  (node-value nd) 'none
	  (node-item nd) (subseq (node-item nd) 0 sub-match))
    child))

;;;
;;; メソッドの定義
;;;

;;; 空か？
(defmethod trie-emptyp ((ts patricia))
  (zerop (patricia-size ts)))

;;; 要素数
(defmethod trie-count ((ts patricia))
  (patricia-size ts))

;;; 空にする
(defmethod trie-clear ((ts patricia))
  (setf (node-child (patricia-root ts)) nil
	(patricia-size ts) 0))

;;; データの挿入
(defmethod trie-set ((ts patricia) seq val)
  (multiple-value-bind
   (nd match sub-match)
   (node-longest-match (patricia-root ts) seq (patricia-elt= ts))
   (cond
    ((zerop sub-match)
     (if (= (length seq) match)
         ;; 終端のチェック
	 (cond
	  ((leaf-p nd)
	   (setf (node-value nd) val)
	   nil)
	  (t
	   (setf (node-value nd) val)
	   (incf (patricia-size ts))
	   t))
       ;; nd に新しい節と葉を追加する
       (progn
	 (insert-new-node-leaf nd (subseq seq match) val)
	 (incf (patricia-size ts))
	 t)))
    (t
     ;; nd を sub-match で分割
     ;; nd - child
     (divide-node nd sub-match)
     (if (= (length seq) match)
         (setf (node-value nd) val)
       (insert-new-node-leaf nd (subseq seq match) val))
     (incf (patricia-size ts))
     t))))

(defmethod trie-push ((ts patricia) seq val)
  (multiple-value-bind
   (nd match sub-match)
   (node-longest-match (patricia-root ts) seq (patricia-elt= ts))
   (cond
    ((zerop sub-match)
     (if (= (length seq) match)
	 (cond
	  ((leaf-p nd)
	   (setf (node-value nd) (cons val (node-value nd)))
	   nil)
	  (t
	   (setf (node-value nd) (list val))
	   (incf (patricia-size ts))
	   t))
       (progn
	 (insert-new-node-leaf nd (subseq seq match) (list val))
	 (incf (patricia-size ts))
	 t)))
    (t
     (divide-node nd sub-match)
     (if (= (length seq) match)
         (setf (node-value nd) (list val))
       (insert-new-node-leaf nd (subseq seq match) (list val)))
     (incf (patricia-size ts))
     t))))

;;; データの探索
(defmethod trie-exists ((ts patricia) seq)
  (multiple-value-bind
   (nd match sub-match)
   (node-longest-match (patricia-root ts) seq (patricia-elt= ts))
   (and (zerop sub-match)
        (= (length seq) match)
	(leaf-p nd))))

(defmethod trie-get ((ts patricia) seq)
  (multiple-value-bind
   (nd match sub-match)
   (node-longest-match (patricia-root ts) seq (patricia-elt= ts))
   (when (and (zerop sub-match)
              (= (length seq) match)
	      (leaf-p nd))
     (node-value nd))))

(defmethod trie-pop ((ts patricia) seq)
  (multiple-value-bind
   (nd match sub-match)
   (node-longest-match (patricia-root ts) seq (patricia-elt= ts))
   (when (and (zerop sub-match)
              (= (length seq) match)
	      (leaf-p nd)
	      (consp (node-value nd)))
     (pop (node-value nd)))))

;;; 削除
(defmethod trie-del ((ts patricia) seq)
  (multiple-value-bind
   (nd match sub-match)
   (node-longest-match (patricia-root ts) seq (patricia-elt= ts))
   (when (and (zerop sub-match)
              (= (length seq) match)
	      (leaf-p nd))
     (setf (node-value nd) 'none)
     (decf (patricia-size ts))
     (when (zerop (patricia-size ts))
       (setf (node-child (patricia-root ts)) nil))
     t)))

;;; 巡回
(defmethod trie-mapc ((ts patricia) func)
  (dolist (x (node-child (patricia-root ts)))
    (node-traverse x func nil)))

;;; 畳み込み
(defmethod trie-fold ((ts patricia) a func)
  (dolist (x (node-child (patricia-root ts)) a)
    (setf a (node-fold x func nil a))))

;;; コピー
(defmethod trie-copy ((ts patricia))
  (let ((new-ts (make-patricia :elt= (patricia-elt= ts))))
    (trie-mapc ts (lambda (k v)
		    (let ((seq-type (if (stringp (car k)) 'string (car (type-of (car k))))))
		      (trie-set new-ts (apply #'concatenate seq-type k) v))))
    new-ts))

;;; 共通接頭辞を持つデータを求める
(defmethod trie-common-prefix ((ts patricia) seq func)
  (multiple-value-bind
   (nd match sub-match)
   (node-longest-match (patricia-root ts) seq (patricia-elt= ts))
   (when (= (length seq) match)
     (let ((seq1 (if (zerop sub-match)
                     (list seq)
                   (list (subseq (node-item nd) sub-match) seq))))
       (when (leaf-p nd)
	 (funcall func seq1 (node-value nd)))
       (dolist (x (node-child nd))
	 (node-traverse x func seq1))))))

;;;
;;; ternary search tree (三分木)
;;;

;;; 節
(defstruct node3
  item left mid right (value 'none))

;;; データ型
(defstruct ternary
  (root (make-node3)) (size 0) (elt= #'char=) (elt< #'char<))

;;; 葉
(defun leaf3-p (nd)
  (and (node3-p nd)
       (not (eq (node3-value nd) 'none))))

;;; 子の探索
(defun child-search-tst (nd x elt= elt<)
  (cond
   ((null nd) nil)
   ((funcall elt= x (node3-item nd)) nd)
   ((funcall elt< x (node3-item nd))
    (child-search-tst (node3-left nd) x elt= elt<))
   (t
    (child-search-tst (node3-right nd) x elt= elt<))))

;;; 子の挿入 (x は node3)
(defun child-insert-tst (nd x elt= elt<)
  (cond
   ((null nd) x)
   ((funcall elt= (node3-item x) (node3-item nd)) nd)
   ((funcall elt< (node3-item x) (node3-item nd))
    (setf (node3-left nd)
	  (child-insert-tst (node3-left nd) x elt= elt<))
    nd)
   (t
    (setf (node3-right nd)
	  (child-insert-tst (node3-right nd) x elt= elt<))
    nd)))

;;; 挿入
(defun node-insert-tst (nd seq i elt= elt<)
  (if (= (length seq) i)
      nd
    (let ((child (child-search-tst (node3-mid nd) (elt seq i) elt= elt<)))
      (unless child
	(setf child (make-node3 :item (elt seq i))
	      (node3-mid nd) (child-insert-tst (node3-mid nd) child elt= elt<)))
      (node-insert-tst child seq (1+ i) elt= elt<))))

;;; 探索
(defun node-search-tst (nd seq i elt= elt<)
  (if (= (length seq) i)
      nd
    (let ((child (child-search-tst (node3-mid nd) (elt seq i) elt= elt<)))
      (when child
	(node-search-tst child seq (1+ i) elt= elt<)))))
	  
;;; 巡回
(defun node-traverse-tst (nd func a)
  (when nd
    (node-traverse-tst (node3-left nd) func a)
    (let ((a1 (cons (node3-item nd) a)))
      (when (leaf3-p nd)
	(funcall func (reverse a1) (node3-value nd)))
      (node-traverse-tst (node3-mid nd) func a1))
    (node-traverse-tst (node3-right nd) func a)))

;;; 畳み込み
(defun node-fold-tst (nd func seq a)
  (when nd
    (setf a (node-fold-tst (node3-left nd) func seq a))
    (let ((seq1 (cons (node3-item nd) seq)))
      (when (leaf3-p nd)
	(setf a (funcall func a (reverse seq1) (node3-value nd))))
      (setf a (node-fold-tst (node3-mid nd) func seq1 a)))
    (setf a (node-fold-tst (node3-right nd) func seq a)))
  a)

;;;
;;; メソッドの定義
;;;

;;; トライは空か
(defmethod trie-emptyp ((ts ternary)) (zerop (ternary-size ts)))

;;; 要素数を求める
(defmethod trie-count ((ts ternary)) (ternary-size ts))

;;; クリア
(defmethod trie-clear ((ts ternary))
  (setf (node3-mid (ternary-root ts)) nil
	(ternary-size ts) 0))

;;; 挿入
(defmethod trie-set ((ts ternary) seq val)
  (let ((nd (node-insert-tst (ternary-root ts) seq 0 (ternary-elt= ts) (ternary-elt< ts))))
    (cond
     ((leaf3-p nd)
      (setf (node3-value nd) val)
      nil)
     (t
      (setf (node3-value nd) val)
      (incf (ternary-size ts))
      t))))

(defmethod trie-push ((ts ternary) seq val)
  (let ((nd (node-insert-tst (ternary-root ts) seq 0 (ternary-elt= ts) (ternary-elt< ts))))
    (cond
     ((leaf3-p nd)
      (setf (node3-value nd) (cons val (node3-value nd)))
      nil)
     (t
      (setf (node3-value nd) (list val))
      (incf (ternary-size ts))
      t))))

;;; 探索
(defmethod trie-exists ((ts ternary) seq)
  (leaf3-p (node-search-tst (ternary-root ts) seq 0 (ternary-elt= ts) (ternary-elt< ts))))

(defmethod trie-get ((ts ternary) seq)
  (let ((nd (node-search-tst (ternary-root ts) seq 0 (ternary-elt= ts) (ternary-elt< ts))))
    (when (leaf3-p nd)
      (node3-value nd))))

;;; 削除
(defmethod trie-del ((ts ternary) seq)
  (let ((nd (node-search-tst (ternary-root ts) seq 0 (ternary-elt= ts) (ternary-elt< ts))))
   (when (leaf3-p nd)
     (setf (node3-value nd) 'none)
     (decf (ternary-size ts))
     (when (zerop (ternary-size ts))
       (setf (node3-mid (ternary-root ts)) nil))       
     t)))

(defmethod trie-pop ((ts ternary) seq)
  (let ((nd (node-search-tst (ternary-root ts) seq 0 (ternary-elt= ts) (ternary-elt< ts))))
    (when (and (leaf3-p nd)
	       (consp (node3-value nd)))
      (pop (node3-value nd)))))

;;; 巡回
(defmethod trie-mapc ((ts ternary) func)
  (node-traverse-tst (node3-mid (ternary-root ts)) func nil))

;;; 畳み込み
(defmethod trie-fold ((ts ternary) a func)
  (node-fold-tst (node3-mid (ternary-root ts)) func nil a))

;;; コピー
(defmethod trie-copy ((ts ternary))
  (let ((new-ts (make-ternary :elt= (ternary-elt= ts) :elt< (ternary-elt< ts))))
    (trie-mapc ts (lambda (k v) (trie-set new-ts k v)))
    new-ts))

;;; 共通接頭辞の取得
(defmethod trie-common-prefix ((ts ternary) seq func)
  (let ((k (coerce seq 'list))
	(nd (node-search-tst (ternary-root ts) seq 0 (ternary-elt= ts) (ternary-elt< ts))))
    (when nd
      (when (leaf3-p nd)
	(funcall func k (node3-value nd)))
      (node-traverse-tst (node3-mid nd) func (reverse k)))))

