;;;
;;; utils.lsp : Common Lisp ユーティリティ
;;;
;;; Copyright (c) 2023 Makoto Hiroi
;;;
;;; Released under the MIT license
;;; https://opensource.org/license/mit/
;;;
(provide :utils)
(defpackage :utils (:use :cl))
(in-package :utils)
(export '(xcons
	  iota
	  tabulate
	  unfold
	  take
	  drop
	  take-while
	  drop-while
	  split-at
	  split-if
	  split-if-not
	  partition
          insert
	  insert-at
	  remove-at
	  selects
	  power-set
          fold-left
	  fold-right
	  scan-left
	  scan-right
	  ;; ベクタ
	  vector-iota
	  vector-tabulate
	  vector-unfold
	  vector-fold-left
	  vector-fold-right
	  vector-map-with-index
	  vector-mapc-with-index
	  vector-swap
	  vector-partition
	  vector-bsearch
	  vector-shuffle
          ))

;;;
;;; リストの生成
;;;
(defun xcons (d a) (cons a d))

(defun tabulate (fn n &key (acc nil) (start 0) (step 1))
  (if (zerop n)
      (nreverse acc)
    (tabulate fn (1- n) :acc (cons (funcall fn start) acc) :start (+ start step) :step step)))

(defun iota (n &key (start 0) (step 1))
  (tabulate #'identity n :start start :step step))

;;; 解きほぐし
(defun unfold (p f g s &optional (a nil))
  (if (funcall p s)
      (nreverse a)
    (unfold p f g (funcall g s) (cons (funcall f s) a))))

;;;
;;; リスト操作
;;;
(defun take (xs n &optional (a nil))
  (if (or (zerop n) (null xs))
      (nreverse a)
    (take (cdr xs) (1- n) (cons (car xs) a))))

(defun drop (xs n)
  (if (or (zerop n) (null xs))
      xs
    (drop (cdr xs) (1- n))))

(defun take-while (pred xs &optional (a nil))
  (if (or (null xs)
	  (not (funcall pred (car xs))))
      (nreverse a)
    (take-while pred (cdr xs) (cons (car xs) a))))

(defun drop-while (pred xs)
  (if (or (null xs)
	  (not (funcall pred (car xs))))
      xs
    (drop-while pred (cdr xs))))

(defun split-at (xs n &optional (a nil))
  (if (or (null xs) (zerop n))
      (values (nreverse a) xs)
    (split-at (cdr xs) (1- n) (cons (car xs) a))))

(defun split-if (pred xs &optional (a nil))
  (if (or (null xs)
	  (not (funcall pred (car xs))))
      (values (nreverse a) xs)
    (split-if pred (cdr xs) (cons (car xs) a))))

(defun split-if-not (pred xs &optional (a nil))
  (if (or (null xs)
	  (funcall pred (car xs)))
      (values (nreverse a) xs)
    (split-if-not pred (cdr xs) (cons (car xs) a))))

(defun partition (pred xs &optional (ys nil) (zs nil))
  (cond
   ((null xs)
    (values (nreverse ys) (nreverse zs)))
   ((funcall pred (car xs))
    (partition pred (cdr xs) (cons (car xs) ys) zs))
   (t
    (partition pred (cdr xs) ys (cons (car xs) zs)))))

(defun insert (xs x pred &optional (a nil))
  (cond
   ((null xs)
    (nreverse (cons x a)))
   ((or (funcall pred x (car xs))
	(not (funcall pred (car xs) x)))
    (nreconc a (cons x xs)))
   (t
    (insert (cdr xs) x pred (cons (car xs) a)))))

(defun insert-at (xs n x &optional (a nil))
  (cond
   ((zerop n)
    (nreconc a (cons x xs)))
   ((null xs)
    (nreverse (cons x a)))
   (t
    (insert-at (cdr xs) (1- n) x (cons (car xs) a)))))

(defun remove-at (xs n &optional (a nil))
  (cond
   ((zerop n)
    (nreconc a (cdr xs)))
   ((null xs)
    (nreverse a))
   (t
    (remove-at (cdr xs) (1- n) (cons (car xs) a)))))


;;; 選んだ要素は CAR に、残りの要素は CDR に格納される
(defun selects (xs)
  (if (null (cdr xs))
      (list (list (car xs)))
    (cons (cons (car xs) (cdr xs))
	  (mapcar (lambda (ys) (cons (car ys) (cons (car xs) (cdr ys))))
		  (selects (cdr xs))))))

;;; べき集合
(defun power-set (ls)
  (if (null ls)
      '(())
    (append (power-set (cdr ls))
            (mapcar (lambda (xs) (cons (car ls) xs))
                    (power-set (cdr ls))))))


;;;
;;; 畳み込み
;;;
(defun fold-left-1 (fn a xs)
  (if (null xs)
      a
    (fold-left-1 fn (funcall fn a (car xs)) (cdr xs))))

(defun fold-left-n (fn a xss)
  (if (member nil xss)
      a
    (fold-left-n fn (apply fn a (mapcar #'car xss)) (mapcar #'cdr xss))))

(defun fold-left (fn a xs &rest args)
  (if (null args)
      (fold-left-1 fn a xs)
    (fold-left-n fn a (cons xs args))))

;;; 末尾再帰ではないので注意
(defun fold-right-1 (fn a xs)
  (if (null xs)
      a
    (funcall fn (fold-right-1 fn a (cdr xs)) (car xs))))

(defun fold-right-n (fn a xss)
  (if (member nil xss)
      a
    (apply fn (fold-right-n fn a (mapcar #'cdr xss)) (mapcar #'car xss))))

(defun fold-right (fn a xs &rest args)
  (if (null args)
      (fold-right-1 fn a xs)
    (fold-right-n fn a (cons xs args))))

(defun scan-left-1 (fn a xs)
  (if (null xs)
      (list a)
    (cons a (scan-left-1 fn (funcall fn a (car xs)) (cdr xs)))))

(defun scan-left-n (fn a xss)
  (if (member nil xss)
      (list a)
    (cons a (scan-left-n fn (apply fn a (mapcar #'car xss)) (mapcar #'cdr xss)))))

(defun scan-left (fn a xs &rest args)
  (if (null args)
      (scan-left-1 fn a xs)
    (scan-left-n fn a (cons xs args))))

(defun scan-right-1 (fn a xs)
  (if (null xs)
      (list a)
    (let ((ys (scan-right-1 fn a (cdr xs))))
      (cons (funcall fn (car ys) (car xs)) ys))))

(defun scan-right-n (fn a xss)
  (if (member nil xss)
      (list a)
    (let ((ys (scan-right-n fn a (mapcar #'cdr xss))))
      (cons (apply fn (car ys) (mapcar #'car xss)) ys))))

(defun scan-right (fn a xs &rest args)
  (if (null args)
      (scan-right-1 fn a xs)
    (scan-right-n fn a (cons xs args))))

;;;
;;; ベクタの操作関数
;;;
(defun vector-tabulate (fn n &key (start 0) (step 1))
  (do ((vec (make-array n))
       (i 0 (1+ i))
       (x start (+ x step)))
      ((>= i n) vec)
      (setf (aref vec i) (funcall fn x))))

(defun vector-iota (n &key (start 0) (step 1))
  (vector-tabulate #'identity n :start start :step step))

(defun vector-unfold (n f g s)
  (do ((vec (make-array n))
       (i 0 (1+ i))
       (s s (funcall g s)))
      ((>= i n) vec)
      (setf (aref vec i) (funcall f s))))

;;;
;;; 添字付きマッピング
;;;
(defun vector-map-with-index-1 (fn vec)
  (do ((xs (make-array (length vec)))
       (i 0 (1+ i)))
      ((>= i (length vec)) xs)
      (setf (aref xs i) (funcall fn i (aref vec i)))))

(defun vector-map-with-index-n (fn vs)
  (do* ((k (apply #'min (mapcar #'length vs)))
	(xs (make-array k))
	(i 0 (1+ i)))
       ((>= i k) xs)
       (setf (aref xs i) (apply fn i (mapcar (lambda (ys) (aref ys i)) vs)))))

(defun vector-map-with-index (fn vec &rest args)
  (if (null args)
      (vector-map-with-index-1 fn vec)
    (vector-map-with-index-n fn (cons vec args))))

(defun vector-mapc-with-index-1 (fn vec)
  (do ((i 0 (1+ i)))
      ((>= i (length vec)))
      (funcall fn i (aref vec i))))

(defun vector-mapc-with-index-n (fn vs)
  (do* ((k (apply #'min (mapcar #'length vs)))
	(i 0 (1+ i)))
       ((>= i k))
       (apply fn i (mapcar (lambda (ys) (aref ys i)) vs))))

(defun vector-mapc-with-index (fn vec &rest args)
  (if (null args)
      (vector-mapc-with-index-1 fn vec)
    (vector-mapc-with-index-n fn (cons vec args))))

;;; 畳み込み
(defun vector-fold-left-1 (fn a vec)
  (dotimes (i (length vec) a)
    (setf a (funcall fn a (aref vec i)))))

(defun vector-fold-left-n (fn a vs)
  (do ((k (apply #'min (mapcar #'length vs)))
       (i 0 (1+ i)))
      ((>= i k) a)
      (setf a (apply fn a (mapcar (lambda (xs) (aref xs i)) vs)))))

(defun vector-fold-left (fn a vec &rest args)
  (if (null args)
      (vector-fold-left-1 fn a vec)
    (vector-fold-left-n fn a (cons vec args))))

(defun vector-fold-right-1 (fn a vec)
  (do ((i (1- (length vec)) (1- i)))
      ((minusp i) a)
      (setf a (funcall fn a (aref vec i)))))

(defun vector-fold-right-n (fn a vs)
  (do* ((k (apply #'min (mapcar #'length vs)))
	(i (1- k) (1- i)))
       ((minusp i) a)
       (setf a (apply fn a (mapcar (lambda (xs) (aref xs i)) vs)))))

(defun vector-fold-right (fn a vec &rest args)
  (if (null args)
      (vector-fold-right-1 fn a vec)
    (vector-fold-right-n fn a (cons vec args))))


;;; 要素の交換
(defun vector-swap (vec i j)
  (psetf (aref vec i) (aref vec j)
	 (aref vec j) (aref vec i))
  vec)

;;; ベクタの分割 (破壊的)
(defun vector-partition (pred vec &key (start 0) (end (length vec)))
  (let ((i start) (j (1- end)))
    (loop
     (do () ((or (<= end i) (not (funcall pred (aref vec i))))) (incf i))
     (do () ((or (> start j) (funcall pred (aref vec j)))) (decf j))
     (cond
      ((>= i j)
       (return (values i vec)))
      (t
       (vector-swap vec i j)
       (incf i)
       (decf j))))))

;;; ベクタの二分探索
(defun bsearch (x xs low high key elt= elt<)
  (when (<= low high)
    (let ((mid (truncate (+ low high) 2)))
     (cond
      ((funcall elt= x (funcall key (aref xs mid))) mid)
      ((funcall elt< x (funcall key (aref xs mid)))
       (bsearch x xs low (1- mid) key elt= elt<))
      (t
       (bsearch x xs (1+ mid) high key elt= elt<))))))

(defun vector-bsearch (x xs &key (start 0) (end (length xs)) (key #'identity) (elt= #'eql) (elt< #'<))
  (bsearch x xs start (1- end) key elt= elt<))

;;; ベクタのシャッフル
(defun vector-shuffle (vec)
  (do ((i (1- (length vec)) (1- i)))
      ((zerop i) vec)
      (let ((j (random (1+ i))))
	(vector-swap vec i j))))

